"""Coordinate systems and function graphing related mobjects.

Modules
=======

.. autosummary::
    :toctree: ../reference

    ~coordinate_systems
    ~functions
    ~number_line
    ~probability
    ~scale
"""
