"""Utilities for Manim tests using `pytest <https://pytest.org>`_.

For more information about Manim testing, see:

-   :doc:`/contributing/development`, specifically the ``Tests`` bullet
    point under :ref:`polishing-changes-and-submitting-a-pull-request`
-   :doc:`/contributing/testing`

.. autosummary::
   :toctree: ../reference

   frames_comparison
   _frames_testers
   _show_diff
   _test_class_makers

"""
