"""Utilities for building the Manim documentation.

For more information about the Manim documentation building, see:

-   :doc:`/contributing/development`, specifically the ``Documentation``
    bullet point under :ref:`polishing-changes-and-submitting-a-pull-request`
-   :doc:`/contributing/docs`

.. autosummary::
   :toctree: ../reference

   autoaliasattr_directive
   autocolor_directive
   manim_directive
   module_parsing

"""
